unit ColorPicker;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, Math;

type
  TRGBQuadArray = array [0..65535] of TRGBQuad;
  pRGBQuadArray = ^TRGBQuadArray;
  TMarkerStyle = (msCircle, msSquare, msCross, msCrossCirc);

  TShColorPicker = class;

  TCustomColorSelect = class(TCustomControl)
    private
      fVal, fSat, fHue: integer;
      fHBmp       : TBitmap;
      Aw          : integer;
      fControl    : TShColorPicker;
      function    YToArrowPos(p: integer): integer;
      function    ArrowPosFromHue(h: integer): integer;
      function    HueFromArrowPos(p: integer): integer;
      procedure   CreateHGradient;
      procedure   SetHue(h: integer);
      procedure   SetSat(s: integer);
      procedure   SetValue(v: integer);
      function    GetSelectedColor: TColor;
      procedure   SetSelectedColor(c: TColor);
    protected
      fArrowPos   : integer;
      fLimit      : integer;
      Mx,My       : integer;
      procedure   Paint; override;
      procedure   PaintMaker;
      procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
    published
      property    Hue: integer read fHue write SetHue default 0;
      property    Saturation: integer read fSat write SetSat default 255;
      property    Value: integer read fVal write SetValue default 255;
      property    SelectedColor: TColor read GetSelectedColor write SetSelectedColor default clRed;
    end;

  TCustomColorPicker = class(TCustomControl)
    private
      fBMP: TBitmap;
      fHue, fSat, fLum: integer;
      fOnChange    : TNotifyEvent;
      procedure   SetHue(h: integer);
      procedure   SetSat(s: integer);
      procedure   SetLum(l: integer);
      procedure   CreateSLGradient;
      procedure   UpdateCoords;
      procedure   DrawMarker(x, y: integer);
      procedure   SelectionChanged(x, y: integer);
      function    GetSelectedColor: TColor;
      procedure   SetSelectedColor(C: TColor);
    protected
      mx, my, mdx, mdy: integer;
      procedure   Paint; override;
      procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
    published
      property    Hue: integer read fHue write SetHue default 0;
      property    Saturation: integer read fSat write SetSat default 0;
      property    Luminance: integer read fLum write SetLum default 255;
      property    SelectedColor: TColor read GetSelectedColor write SetSelectedColor;
      property    OnChange: TNotifyEvent read fOnChange write fOnChange;
    end;

  TShColorPicker = class(TCustomControl)
    private
      function     GetSelectedColor: TColor;
      procedure    SetSelectedColor(C: TColor);
    protected
      procedure   WMSize(var Message: TWMSize); message WM_SIZE;
    public
      fColorSelect : TCustomColorSelect;
      fColorPicker : TCustomColorPicker;
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
    published
      property    SelectedColor: TColor read GetSelectedColor write SetSelectedColor;
    end;

implementation

var //set these variables to your needs, e.g. 360, 255, 255
  MaxHue: integer = 239;
  MaxSat: integer = 240;
  MaxLum: integer = 240;

procedure Clamp(var Input: integer; Min, Max: integer);
begin
  if (Input < Min) then Input := Min;
  if (Input > Max) then Input := Max;
end;

function RGBtoRGBTriple(R, G, B: byte): TRGBTriple;
begin
  with Result do begin
    rgbtRed := R;
    rgbtGreen := G;
    rgbtBlue := B;
  end
end;

function RGBtoRGBQuad(R, G, B: byte): TRGBQuad; overload;
begin
  with Result do begin
    rgbRed := R;
    rgbGreen := G;
    rgbBlue := B;
    rgbReserved := 0;
  end
end;

function RGBToRGBQuad(c: TColor): TRGBQuad; overload;
begin
  with Result do begin
    rgbRed := GetRValue(c);
    rgbGreen := GetGValue(c);
    rgbBlue := GetBValue(c);
    rgbReserved := 0
  end;
end;

procedure RGBTripleToHSL(RGBTriple: TRGBTriple; var h, s, l: integer);

 function RGBMaxValue(RGB: TRGBTriple): byte;
 begin
  Result  := RGB.rgbtRed;
  if (Result < RGB.rgbtGreen) then Result := RGB.rgbtGreen;
  if (Result < RGB.rgbtBlue) then Result := RGB.rgbtBlue;
 end;

 function RGBMinValue(RGB: TRGBTriple) : byte;
 begin
  Result  := RGB.rgbtRed;
  if (Result > RGB.rgbtGreen) then Result := RGB.rgbtGreen;
  if (Result > RGB.rgbtBlue) then Result := RGB.rgbtBlue;
 end;

var
 Delta, Min: byte;

begin
 L     := RGBMaxValue(RGBTriple);
 Min   := RGBMinValue(RGBTriple);
 Delta := L-Min;
 if (L = Min) then
  begin
   H := 0;
   S := 0;
  end
 else
  begin
   S := MulDiv(Delta, 255, L);
   with RGBTriple do
    begin
     if (rgbtRed = L) then
      H := MulDiv(60, rgbtGreen-rgbtBlue, Delta)
     else
      if (rgbtGreen = L) then
       H := MulDiv(60, rgbtBlue-rgbtRed,   Delta) + 120
      else
       if (rgbtBlue = L) then
        H := MulDiv(60, rgbtRed-rgbtGreen,  Delta) + 240;
     if (H < 0) then H := H + 360;
    end;
  end;
end;

function HSLToRGBQuad(H, S, L: integer): TRGBQuad;
const
  Divisor = 255*60;
var
  hTemp, f, LS, p, q, r: integer;
begin
  Clamp(H, 0, MaxHue);
  Clamp(S, 0, MaxSat);
  Clamp(L, 0, MaxLum);
  if (S = 0) then
    Result := RGBToRGBQuad(L, L, L)
  else begin
    hTemp := H mod MaxHue;
    f := hTemp mod 60;
    hTemp := hTemp div 60;
    LS := L*S;
    p := L - LS div MaxLum;
    q := L - (LS*f) div Divisor;
    r := L - (LS*(60 - f)) div Divisor;
    case hTemp of
      0: Result := RGBToRGBQuad(L, r, p);
      1: Result := RGBToRGBQuad(q, L, p);
      2: Result := RGBToRGBQuad(p, L, r);
      3: Result := RGBToRGBQuad(p, q, L);
      4: Result := RGBToRGBQuad(r, p, L);
      5: Result := RGBToRGBQuad(L, p, q);
    else
      Result  := RGBToRGBQuad(0, 0, 0);
    end;
  end;
end;

function HSLToRGBTriple(H, S, L: integer): TRGBTriple;
const
  Divisor = 255*60;
var
  hTemp, f, LS, p, q, r: integer;
begin
  Clamp(H, 0, MaxHue);
  Clamp(S, 0, MaxSat);
  Clamp(L, 0, MaxLum);
  if (S = 0) then
    Result := RGBToRGBTriple(L, L, L)
  else begin
    hTemp := H mod MaxHue;
    f := hTemp mod 60;
    hTemp := hTemp div 60;
    LS := L*S;
    p := L - LS div MaxLum;
    q := L - (LS*f) div Divisor;
    r := L - (LS*(60 - f)) div Divisor;
    case hTemp of
      0: Result := RGBToRGBTriple(L, r, p);
      1: Result := RGBToRGBTriple(q, L, p);
      2: Result := RGBToRGBTriple(p, L, r);
      3: Result := RGBToRGBTriple(p, q, L);
      4: Result := RGBToRGBTriple(r, p, L);
      5: Result := RGBToRGBTriple(L, p, q);
    else
      Result  := RGBToRGBTriple(0, 0, 0);
    end;
  end;
end;

function HSVtoRGBTriple(H, S, V: integer): TRGBTriple;
const
 divisor: integer = 255*60;
var
 f, hTemp, p, q, t, VS: integer;
begin
 if H > 360 then H := H - 360;
 if H < 0 then H := H + 360;
 if s = 0 then
  Result := RGBtoRGBTriple(V, V, V)
 else
  begin
   if H = 360 then hTemp := 0 else hTemp := H;
   f := hTemp mod 60;
   hTemp := hTemp div 60;
   VS := V*S;
   p := V - VS div 255;
   q := V - (VS*f) div divisor;
   t := V - (VS*(60 - f)) div divisor;
   case hTemp of
    0: Result := RGBtoRGBTriple(V, t, p);
    1: Result := RGBtoRGBTriple(q, V, p);
    2: Result := RGBtoRGBTriple(p, V, t);
    3: Result := RGBtoRGBTriple(p, q, V);
    4: Result := RGBtoRGBTriple(t, p, V);
    5: Result := RGBtoRGBTriple(V, p, q);
   else Result := RGBtoRGBTriple(0,0,0)
   end;
  end;
end;

function RGBTripleToTColor(RGBTriple: TRGBTriple): TColor;
begin
  Result := RGBTriple.rgbtBlue shl 16 + RGBTriple.rgbtGreen shl 8 + RGBTriple.rgbtRed;
end;

function RGBTripleToColor(Triple: TRGBTriple): TColor;
begin
 Result := TColor(RGB(Triple.rgbtRed, Triple.rgbtGreen, Triple.rgbtBlue));
end;

function HSVtoColor(H, S, V: integer): TColor;
begin
 Result := RGBTripleToColor(HSVtoRGBTriple(H, S, V));
end;

function MinIntValue(const Data: array of Integer): Integer;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

function MaxIntValue(const Data: array of Integer): Integer;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;

procedure RGBToHSV(R, G, B: integer; var H, S, V: integer);
var
 Delta, Min, H1, S1: real;
begin
 h1 := h;
 Min := MinIntValue([R, G, B]);
 V := MaxIntValue([R, G, B]);
 Delta := V - Min;
 if V =  0.0 then S1 := 0 else S1 := Delta / V;
 if S1  = 0.0 then
  H1 := 0
 else
  begin
   if R = V then
    H1 := 60.0 * (G - B) / Delta
   else
    if G = V then
     H1 := 120.0 + 60.0 * (B - R) / Delta
    else
     if B = V then
      H1 := 240.0 + 60.0 * (R - G) / Delta;
   if H1 < 0.0 then H1 := H1 + 360.0;
  end;
 h := round(h1);
 s := round(s1*255);
end;

procedure DrawSelCirc(x, y: integer; Canvas: TCanvas);
var
  R: TRect;
begin
  R := Rect(x - 5, y - 5, x + 5, y + 5);
  Canvas.Brush.Style := bsClear;
  Canvas.Pen.Mode := pmNot;
  Canvas.Ellipse(R);
  Canvas.Pen.Mode := pmCopy;
  Canvas.Brush.Style := bsSolid;
end;

procedure RGBtoHSLRange(RGB: TColor; var H1, S1, L1 : integer);
var
  R, G, B, D, Cmax, Cmin, h, s, l: double;
begin
 R := GetRValue (RGB) / 255;
 G := GetGValue (RGB) / 255;
 B := GetBValue (RGB) / 255;
 Cmax := Max (R, Max (G, B));
 Cmin := Min (R, Min (G, B));
 L := (Cmax + Cmin) / 2;
 if Cmax = Cmin then
  begin
   H := 0;
   S := 0;
  end
 else
  begin
   D := Cmax - Cmin;
   //calc L
   if L < 0.5 then
    S := D / (Cmax + Cmin)
   else
    S := D / (2 - Cmax - Cmin);
   //calc H
   if R = Cmax then
    H := (G - B) / D
   else
    if G = Cmax then
     H  := 2 + (B - R) /D
    else
     H := 4 + (R - G) / D;
   H := H / 6;
   if H < 0 then
    H := H + 1;
  end;
 H1 := round (H * MaxHue);
 S1 := round (S * MaxSat);
 L1 := round (L * MaxLum);
end;

function GetHValue(AColor: TColor): integer;
var
 d, h: integer;
begin
 RGBToHSLRange(AColor, h, d, d);
 Result := h;
end;

//######################################################################

constructor TCustomColorSelect.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  DoubleBuffered:=true;
  fControl := TShColorPicker(AOwner);
  fHBmp := TBitmap.Create;
  fHBmp.PixelFormat := pf32bit;
  Aw        := 4;
  fLimit    := 7;
  fSat      := 255;
  fVal      := 255;
  fArrowPos := ArrowPosFromHue(0);
  SetHue(0);
  CreateHGradient;
end;

destructor TCustomColorSelect.Destroy;
begin
  fHBmp.Free;
  inherited;
end;

procedure TCustomColorSelect.CreateHGradient;
var
  i,j: integer;
  row: pRGBQuadArray;
begin
  if fHBmp = nil then begin
    fHBmp := TBitmap.Create;
    fHBmp.PixelFormat := pf32bit;
  end;
  fHBmp.width := 12;
  fHBmp.height := 360;
  for i := 0 to 359 do begin
    row := FHBmp.ScanLine[i];
    for j := 0 to 11 do
      row[j] := RGBtoRGBQuad(HSVtoColor(i, fSat, fVal));
  end;
end;

procedure TCustomColorSelect.SetValue(v: integer);
begin
  if v < 0 then v := 0;
  if v > 255 then v := 255;
  if fVal <> v then begin
    fVal := v;
    CreateHGradient;
    Invalidate;
  end;
end;

procedure TCustomColorSelect.SetHue(h: integer);
begin
  if h > 360 then h := 360;
  if h < 0 then h := 0;
  if fHue <> h then begin
    fHue := h;
    fArrowPos := ArrowPosFromHue(h);
    Invalidate;
  end;
end;

procedure TCustomColorSelect.SetSat(s: integer);
begin
  if s > 255 then s := 255;
  if s < 0 then s := 0;
  if fSat <> s then begin
    fSat := s;
    CreateHGradient;
    Invalidate;
  end;
end;

function TCustomColorSelect.GetSelectedColor: TColor;
begin
  Result := HSVtoColor(fHue, fSat, fVal);
end;

procedure TCustomColorSelect.SetSelectedColor(c: TColor);
var
  h, s, v: integer;
begin
  RGBToHSV(GetRValue(c), GetGValue(c), GetBValue(c), h, s, v);
  SetHue(h);
  SetSat(s);
  SetValue(v);
end;

function TCustomColorSelect.YToArrowPos(p: integer): integer;
var
  pos: integer;
begin
  pos := p - Aw;
  if pos < 0 then pos := 0;
  if pos > Height - Aw*2 - 1 then pos := Height - Aw*2 - 1;
  Result := pos;
end;

function TCustomColorSelect.ArrowPosFromHue(h: integer): integer;
var
  a: integer;
begin
  a := Round(((Height - 12)/360)*h);
  if a > Height - fLimit then a := Height - fLimit;
  if a < 0 then a := 0;
  Result := a;
end;

function TCustomColorSelect.HueFromArrowPos(p: integer): integer;
var
  r: integer;
begin
  r := Round(p/((Height - 12)/360));
  if r < 0 then r := 0;
  if r > 360 then r := 360;
  Result := r;
end;

procedure TCustomColorSelect.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if ssLeft in shift then begin
    fArrowPos := YToArrowPos(y);
    fHue:=HueFromArrowPos(fArrowPos);
    fControl.fColorPicker.Hue:=fHue;
    //Invalidate;
    PaintMaker;
    My:=Y;
  end;
  inherited;
end;

procedure TCustomColorSelect.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then Exit;
  SetFocus;
  fArrowPos := YToArrowPos(y);
  fHue:=HueFromArrowPos(fArrowPos);
  fControl.fColorPicker.Hue:=fHue;
  Invalidate;
  My:=Y;
  inherited;
end;

procedure TCustomColorSelect.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then Exit;
  fArrowPos := YToArrowPos(y);
  fHue:=HueFromArrowPos(fArrowPos);
  fControl.fColorPicker.Hue:=fHue;
  Invalidate;
  My:=Y;
  inherited;
end;

procedure TCustomColorSelect.PaintMaker;
var
  X,Y : integer;
begin
  X:=7;
  Y:=fArrowPos+Aw;
  with Canvas do begin
    Brush.Color := Color;
    Canvas.FillRect(Rect(x-7,My-Aw,x+1,My+Aw+1));
    Brush.Color := clWindow;
    Pen.Color := clBtnShadow;
    Polygon([
      Point(x, y),
      Point(x - 3, y -+ 3),
      Point(x - 5, y - 3),
      Point(x - 6, y - 2),
      Point(x - 6, y + 2),
      Point(x - 5, y + 3),
      Point(x - 3, y + 3)
    ]);
    X:=X+21;
    Brush.Color := Color;
    Canvas.FillRect(Rect(x,My-Aw,x+7,My+Aw+1));
    Brush.Color := clWindow;
    Pen.Color := clBtnShadow;
    Polygon([
      Point(x, y),
      Point(x + 3, y - 3),
      Point(x + 5, y - 3),
      Point(x + 6, y - 2),
      Point(x + 6, y + 2),
      Point(x + 5, y + 3),
      Point(x + 3, y + 3)
    ]);
  end;
end;

procedure TCustomColorSelect.Paint;
begin
  Canvas.StretchDraw(Rect(8,Aw,28,Height-Aw), fHBMP);
  PaintMaker;
end;

//######################################################################

constructor TCustomColorPicker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  DoubleBuffered := True;
  fBMP := TBitmap.Create;
  fBMP.PixelFormat := pf32bit;
  fBMP.SetSize(256, 256);
  MaxHue := 360;
  MaxSat := 255;
  MaxLum := 255;
  FHue := 0;
  FSat := 0;
  FLum := 255;
  Left:=2;
  Top:=2;
  CreateSLGradient;
  DoubleBuffered:=true;
  UpdateCoords;
end;

destructor TCustomColorPicker.Destroy;
begin
  fBMP.Free;
  inherited;
end;

procedure TCustomColorPicker.CreateSLGradient;
var
 x, y, skip: integer;
 row: pRGBQuadArray;
begin
  if fBMP = nil then begin
    fBMP := TBitmap.Create;
    fBMP.PixelFormat := pf32bit;
    fBMP.Width := 256;
    fBMP.Height := 256;
  end;
  row := fBMP.ScanLine[0];
  skip := integer(FBMP.ScanLine[1]) - Integer(row);
  for y := 0 to 255 do begin
    for x := 0 to 255 do
      row[x] := HSLtoRGBQuad(FHue, x, 255 - y);
    row := pRGBQuadArray(Integer(row) + skip);
  end;
end;

procedure TCustomColorPicker.SetHue(h: integer);
begin
  if h > 360 then h := 360;
  if h < 0 then h := 0;
  if fHue <> h then begin
    fHue := h;
    CreateSLGradient;
    UpdateCoords;
    Invalidate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TCustomColorPicker.SetSat(s: integer);
begin
  if s > 255 then s := 255;
  if s < 0 then s := 0;
  if FSat <> s then begin
    fSat := s;
    UpdateCoords;
    Invalidate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TCustomColorPicker.SetLum(l: integer);
begin
  if l > 255 then l := 255;
  if l < 0 then l := 0;
  if fLum <> l then begin
    fLum := l;
    UpdateCoords;
    Invalidate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TCustomColorPicker.SelectionChanged(x, y: integer);
begin
  SetSat(MulDiv(255, x, Width - 1));
  SetLum(MulDiv(255, Height - y -1, Height - 1));
end;

procedure TCustomColorPicker.SetSelectedColor(c: TColor);
var
  h, s, l: integer;
begin
  RGBTripleToHSL(RGBtoRGBTriple(GetRValue(c), GetGValue(c), GetBValue(c)), h, s, l);
  SetHue(h);
  SetSat(s);
  SetLum(l);
end;

function TCustomColorPicker.GetSelectedColor: TColor;
var
  triple: TRGBTriple;
begin
 triple := HSLToRGBTriple(FHue, FSat, FLum);
 Result := RGBTripleToTColor(triple)
end;

procedure TCustomColorPicker.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if csDesigning in ComponentState then Exit;
  if (ssLeft in Shift) then begin
    mdx := x;
    mdy := y;
    SelectionChanged(X, Y);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TCustomColorPicker.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if csDesigning in ComponentState then Exit;
  if (Button = mbLeft) and PtInRect(ClientRect, Point(x, y)) then begin
    mdx := x;
    mdy := y;
    SelectionChanged(X, Y);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TCustomColorPicker.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if csDesigning in ComponentState then Exit;
  if (Button = mbLeft) and PtInRect(ClientRect, Point(x, y)) then begin
    mdx := x;
    mdy := y;
    SelectionChanged(X, Y);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
  SetFocus;
end;

procedure TCustomColorPicker.UpdateCoords;
begin
  mdx := MulDiv(FSat, Width, 255);
  mdy := MulDiv(255-FLum, Height, 255);
end;

procedure TCustomColorPicker.DrawMarker(x, y: integer);
begin
  DrawSelCirc(x, y, Canvas);
end;

procedure TCustomColorPicker.Paint;
begin
  Canvas.StretchDraw(Rect(0,0,Width,Height), fBMP);
  DrawMarker(mdx, mdy);
end;

//######################################################################

constructor TShColorPicker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 220;
  Height:= 180;

  fColorSelect        := TCustomColorSelect.Create(Self);
  fColorSelect.Parent := Self;
  fColorSelect.Height := Height;
  fColorSelect.Width  := 35;
  fColorSelect.Left   := Width-fColorSelect.Width-4;

  fColorPicker        := TCustomColorPicker.Create(Self);
  fColorPicker.Parent := Self;
  fColorPicker.Top    := 4;
  fColorPicker.Left   := 4;
  fColorPicker.Height := Height-8;
  fColorPicker.Width  := fColorSelect.Left-4;

end;

destructor TShColorPicker.Destroy;
begin
  fColorSelect.Free;
  fColorPicker.Free;
  inherited;
end;

procedure TShColorPicker.WMSize(var Message: TWMSize);
begin
  fColorSelect.Height := Height;
  fColorSelect.Width  := 35;
  fColorSelect.Left   := Width-fColorSelect.Width-4;
  fColorPicker.Top    := 4;
  fColorPicker.Left   := 4;
  fColorPicker.Height := Height-8;
  fColorPicker.Width  := fColorSelect.Left-4;
  inherited;
end;

function TShColorPicker.GetSelectedColor: TColor;
begin
  Result := fColorPicker.SelectedColor;
end;

procedure TShColorPicker.SetSelectedColor(c: TColor);
begin
  if C<>fColorPicker.SelectedColor then begin
    fColorSelect.Hue := GetHValue(c);
    fColorPicker.SelectedColor:=C;
  end;
end;

end.
